# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np
from .EventSelectionModel import EventSelectionModel
  
MAX_TRADE = 10000    
    
class TokenDeltaModel():
    
    def __init__(self, max_trade = 100, shape=1, scale=1):
        self.__shape = shape
        self.__scale = scale
        self.__max_trade = max_trade

    def apply(self, n = 1):
        
        if(n == 1):
            rval = self.delta(self.__max_trade)
            return min(rval, self.__max_trade)  
        else:
    
            res = []
            for k in range(n):
                rval = self.delta(self.__max_trade)
                res.append(min(rval,self.__max_trade))
                
            return res  
        
    def set_param(self, max_trade = 100, shape=1, scale=1):
        self.__shape = shape
        self.__scale = scale
        self.__max_trade = max_trade        
        
    def delta(self, p=1):
        self.__scale = self.__max_trade/5
        return self.add_sub(p)*np.random.gamma(self.__shape, self.__scale)   
    
    def add_sub(self, p):
        return 1 if bool(EventSelectionModel().bi_select(p)) else -1     


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...erc import ERC20
from ..exchg import UniswapExchange
from ..quote import LPQuote
from ...utils.data import UniswapExchangeData
from ...utils.data import FactoryData
import math

MINIMUM_LIQUIDITY = 1e-15

class ChildUniswapExchange(UniswapExchange):
    
    
    def __init__(self, factory_struct: FactoryData, exchg_struct: UniswapExchangeData) -> None:
        super().__init__(factory_struct, exchg_struct)
        self.hybrid_supply = 0
        self.hybrid_liquidity_providers = {}

        
    def info(self):
        print(f"Exchange {self.name} ({self.symbol})")
        print(f"Coins: {self.token0}/{self.token1}")
        print(f"Reserves: {self.token0} = {self.reserve0} | {self.token1} = {self.reserve1}")
        print(f"Liquidity: {self.total_supply}")  
        print(f"Hybrid Liquidity: {self.hybrid_supply} \n") 
        
    
    def remove_liquidity(self, to, liquidity, amount0_min, amount1_min):
        
        """ remove_liquidity

            Remove liquidity from both coins in the pair based on lp amount
                
            Parameters
            -------
            to : str
                receiving user address  
            liquidity : float
                lp amount to removed                 
            amount0_min : float
                Minimum amount of A  
            amount0_min : float
                Minimum amount of B 
                
            Returns
            -------
            amount0 : float
                removed liquidity from reserve0    
            amount1 : float
                removed liquidity from reserve1                    
        """         
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), "Error"

        balance0 = tokens.get(self.token0).token_total
        balance1 = tokens.get(self.token1).token_total
        total_liquidity = self.liquidity_providers[to]
        if liquidity >= total_liquidity:
            liquidity = total_liquidity

        amount0 = liquidity * balance0 / self.total_supply      # using balances ensures pro-rata distribution
        amount1 = liquidity * balance1 / self.total_supply      # using balances ensures pro-rata distribution
    
        if not (round(amount0,5) >= round(amount0_min,5)):
            print('amount0 {} amount0_min {}'.format(round(amount0,5), round(amount0_min,5)))        
        
        assert amount0 > 0 and amount1 > 0, 'UniswapV2: INSUFFICIENT_LIQUIDITY_BURNED'
        assert round(amount0,5) >= round(amount0_min,5), 'UniswapV2Router: INSUFFICIENT_A_AMOUNT'
        assert round(amount1,5) >= round(amount1_min,5), 'UniswapV2Router: INSUFFICIENT_B_AMOUNT'
        
        ibalance0, ibalance1 = self._get_iamounts(balance0, balance1)
        iamount0, iamount1 = self._get_iamounts(amount0, amount1)
        
        hybrid_liquidity = min(self.hybrid_supply*iamount0/ibalance0,
                                self.hybrid_supply*iamount1/ibalance1)
        
        self._burn_hybrid(to, hybrid_liquidity)
        self.burn(to, liquidity, amount0, amount1)
        return amount0, amount1    
    
    def mint(self, to, _amount0, _amount1):
        
        """ mint

            Mint new liquidity based on amounts on each coin in the pair
                
            Parameters
            -------
            to : str
                receiving user address      
            _amount0 : float
                desired amount of A      
            _amount1 : float
                desired amount of B                   
        """          
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), "Error"

        balance0 = tokens.get(self.token0).token_total
        balance1 = tokens.get(self.token1).token_total

        amount0 = balance0 - self.reserve0
        amount1 = balance1 - self.reserve1
        
        assert round(amount0,3) == round(_amount0,3)
        assert round(amount1,3) == round(_amount1,3)
    
        liquidity = self._calc_liquidity(amount0, amount1, 'standard')
        hybrid_liquidity = self._calc_liquidity(amount0, amount1, 'hybrid')

        assert liquidity > 0, 'UniswapV2: INSUFFICIENT_LIQUIDITY_MINTED'
        
        self._update(balance0, balance1)
        self._mint(to, liquidity)  
        self._mint_hybrid(to, hybrid_liquidity)
        
    def _calc_liquidity(self, amount0, amount1, liq_type):

        # keeping track of the liquidity providers
        if(liq_type == 'standard'):
            res0 = self.reserve0
            res1 = self.reserve1
            tot_supply = self.total_supply 
        elif(liq_type == 'hybrid'):
            res0, res1 = self._get_iamounts(self.reserve0, self.reserve1)    
            amount0, amount1 = self._get_iamounts(amount0, amount1)   
            tot_supply = self.hybrid_supply
        
        if tot_supply != 0:
            liquidity = min(
                amount0 * tot_supply / res0,
                amount1 * tot_supply / res1
            )
        else:
            liquidity = math.sqrt(amount0 * amount1) - MINIMUM_LIQUIDITY
            if(liq_type == 'standard'):
                self._mint("0", MINIMUM_LIQUIDITY)
            elif(liq_type == 'hybrid'):    
                self._mint_hybrid("0", MINIMUM_LIQUIDITY)                
                
        return liquidity  
    
    def _get_iamounts(self, amt0, amt1):

        tokens = self.factory.token_from_exchange[self.name]
        parent_lp = self.factory.parent_lp
        x_tkn = tokens.get(self.token0) 
        y_tkn = tokens.get(self.token1)

        if(x_tkn.type == 'index'):
            iamt0 = amt0
        else:    
            iamt0 = LPQuote().get_lp_from_amount(parent_lp, x_tkn, amt0)

        if(y_tkn.type == 'index'):
            iamt1 = amt1
        else:    
            iamt1 = LPQuote().get_lp_from_amount(parent_lp, y_tkn, amt1)

        return iamt0, iamt1  
        
    def _mint_hybrid(self, to, value):
        
        """ _mint_hybrid

            Update reserve amounts for both coins in the pair
                
            Parameters
            -------   
            to : str
                receiving user address       
            value : float
                amount of new liquidity                  
        """          
        
        if self.hybrid_liquidity_providers.get(to):
            self.hybrid_liquidity_providers[to] += value
        else:
            self.hybrid_liquidity_providers[to] = value

        self.hybrid_supply += value           
                   
    def _burn_hybrid(self, to, value):
        
        """ _burn

            Burn liquidity from both coins in the pair based on lp amount
                
            Parameters
            -------
            to : str
               receiving user address  
            value : float
                amount of liquidity to be burned                           
        """            
        
        available_liquidity = self.hybrid_liquidity_providers.get(to)
        self.hybrid_liquidity_providers[to] = available_liquidity - value
        self.hybrid_supply -= value          


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...erc import ERC20
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import TickMath

class RebaseIndexToken():
    
    """ 
        Determine rebase amount of index token given a certain amount of liquidity from LP (inverse of SettlementLPToken)  
    """      

    
    def __init__(self):
        pass
     
    def apply(self, lp, tkn, liq_amt, lwr_tick = None, upr_tick = None): 
       
        """ apply

            Apply rebase token calculation settlement; given liquidity amount, what is the reserve amount
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair       
            liq_amt: float
                Liquidity amount to be priced in in terms of reserve token                 

            Returns
            -----------------
            rebase_amount: float
                Reserve token amount, given liquidity amount
                   
        """        
        if(lp.version == UniswapExchangeData.VERSION_V2):
            settlement = self.calc_univ2_tkn_settlement(lp, tkn, liq_amt)
        elif(lp.version == UniswapExchangeData.VERSION_V3):   
            settlement = self.calc_univ3_tkn_settlement(lp, tkn, liq_amt, lwr_tick, upr_tick)
        
        return settlement
    
    def calc_univ2_tkn_settlement(self, lp, token_in, dL):
            
        (x, y) = self.get_reserves(lp, token_in)
        L = lp.get_liquidity()
        a0 = dL*x/L
        a1 = dL*y/L
        gamma = 997/1000

        dy1 = a1
        dy2 = gamma*a0*(y - a1)/(x - a0 + gamma*a0)
        itkn_amt = dy1 + dy2

        return itkn_amt if itkn_amt > 0 else 0  

    def calc_univ3_tkn_settlement(self, lp, token_in, dL, lwr_tick, upr_tick):
                
        L = lp.get_liquidity()
        L_diff = (L - dL) 
        if(token_in.token_name == lp.token0):
            sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/2**96
            sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/2**96 
            sqrtp_cur = lp.slot0.sqrtPriceX96/2**96 
            dPy = (sqrtp_cur - sqrtp_pa)
            dPx = (1/sqrtp_cur - 1/sqrtp_pb)  
            dx = dL*dPx
            dy = dL*dPy    
            sqrtp_next = sqrtp_cur + (997*dy)/(L_diff*1000) 
            itkn_amt = dx + L_diff * (1/sqrtp_cur - 1/sqrtp_next)
        elif(token_in.token_name == lp.token1):
            sqrtp_cur = 2**96/lp.slot0.sqrtPriceX96
            sqrtp_pa = 2**96/TickMath.getSqrtRatioAtTick(lwr_tick)
            sqrtp_pb = 2**96/TickMath.getSqrtRatioAtTick(upr_tick)
            dPy = (1/sqrtp_cur - 1/sqrtp_pa)
            dPx = (sqrtp_cur - sqrtp_pb)
            dx = dL*dPx
            dy = dL*dPy
            sqrtp_next = sqrtp_cur + (997*dx)/(L_diff*1000) 
            itkn_amt = dy + L_diff * (1/sqrtp_cur - 1/sqrtp_next)

        return itkn_amt if itkn_amt > 0 else 0 
    

    def get_reserves(self, lp, token_in):
        tokens = lp.factory.token_from_exchange[lp.name]
        if(token_in.token_name == lp.token1):
            x = lp.get_reserve(tokens[lp.token0])
            y = lp.get_reserve(tokens[lp.token1])
        else: 
            x = lp.get_reserve(tokens[lp.token1])
            y = lp.get_reserve(tokens[lp.token0])
        return (x, y)          
            


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

GWEI_PRECISION = 18

class ERC20:
    
    """ ERC20 token

        Parameters
        -----------------
        token_name : str
            Token name 
        token_addr : str
            Token address  
        token_total : float
            Token holdings 
    """   
    
    def __init__(self, name: str, addr: str, decimal: int = None):
        self.token_name = name
        self.token_addr = addr
        self.token_total = 0
        self.token_decimal = GWEI_PRECISION if decimal == None else decimal
        self.type = 'standard'

    def deposit(self, _from, value):
        
        """ deposit

            Deposit token
                
            Parameters
            -----------------
            _from : str
                user address   
            value : str
                delta to add to total                
        """           
        
        self.token_total += value

    def transfer(self, _to, value):
        
        """ transfer

            Transfer token
                
            Parameters
            -----------------
            _to : str
                user address   
            value : str
                delta to remove from total                
        """         
        
        self.token_total -= value
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np

class EventSelectionModel():
    
    FIRST = 0 
    SECOND = 1 
    THIRD = 1     

    def bi_select(self, p):  
        p = max(min(p, 1), 0)
        return np.random.choice(2, 1, p=[1-p,p])[0]
    
    def tri_select(self, p1, p2):     
        p3 = 1 - p1 - p2
        return np.random.choice(3, 1, p=[p1, p2, p3])[0]    


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import math
from scipy import optimize
from ..Process import Process
from ..liquidity import AddLiquidity
from ..swap import Swap
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Helper
from ...utils.tools.v3 import TickMath

class SwapDeposit(Process):
    
    """ Process to swap approx. half of single token X for token Y (and vice verse) and deposit proceeds
        plus remaining other approximated half

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts        
    """       

    def __init__(self, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
            
    def apply(self, lp, token_in, user_nm, amount_in, lwr_tick = None, upr_tick = None):   
        
        """ apply

            Swap approx. half of single token X for token Y (and vice verse) and deposit proceeds
            plus remaining other approximated half
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_in : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_in : float
                token amount to be swap 
            lwr_tick : int
                lower tick of the position in which to add liquidity   
            upr_tick : int
                upper tick of the position in which to add liquidity   
                
            Returns
            -------
            (amount_in, amount_out) : float, float
                token swap amounts                
        """          
        
        amount_in = tDel.delta() if amount_in == None else amount_in    
        trading_token = self.get_trading_token(lp, token_in)

        # Step 2: deposit   
        if(lp.version == UniswapExchangeData.VERSION_V2):

            # Step 1: swap 
            p_in = self._calc_univ2_deposit_portion(lp, token_in, amount_in)
            amount_out = Swap().apply(lp, token_in, user_nm, p_in*amount_in)

            # Step 2: deposit   
            if(token_in.token_name == lp.token1):
                balance0 = amount_out 
                balance1 = lp.quote(balance0, lp.reserve0, lp.reserve1)
                deposited = balance1 + p_in*amount_in
            elif(token_in.token_name == lp.token0):
                balance1 = amount_out
                balance0 = lp.quote(balance1, lp.reserve1, lp.reserve0)
                deposited = balance0 + p_in*amount_in
            lp.add_liquidity(user_nm, balance0, balance1, balance0, balance1) 
            
        elif(lp.version == UniswapExchangeData.VERSION_V3):  

            # Step 1: swap 
            p_in = self._calc_univ3_deposit_portion(lp, token_in, amount_in, lwr_tick, upr_tick)
            amount_out = Swap().apply(lp, token_in, user_nm, p_in*amount_in)
            
            sqrt_P = lp.slot0.sqrtPriceX96/2**96
            tokens = lp.factory.token_from_exchange[lp.name] 

            # Step 2: deposit 
            if(token_in.token_name == lp.token0):
                balance1 = abs(amount_out) 
                liq = UniV3Helper().calc_Ly(sqrt_P, balance1, lwr_tick, upr_tick)
                deposited = liq/sqrt_P + p_in*amount_in
                                
            elif(token_in.token_name == lp.token1): 
                balance0 = abs(amount_out) 
                liq = UniV3Helper().calc_Lx(sqrt_P, balance0, lwr_tick, upr_tick)
                deposited = liq*sqrt_P + p_in*amount_in
                
            lp.mint(user_nm, lwr_tick, upr_tick, liq)  
                            
        return deposited  

    def get_trading_token(self, lp, token):
        
        """ get_trading_token

            Get opposing token from specified token
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token      
                
            Returns
            -------
            trading_token : ERC20 
                opposing ERC20 token                   
        """         
        
        tokens = lp.factory.token_from_exchange[lp.name]
        trading_token = tokens[lp.token1] if token.token_name == lp.token0 else tokens[lp.token0]
        return trading_token       


    def _calc_univ3_deposit_portion(self, lp, tkn, amt_tkn_in, lwr_tick, upr_tick):
        bnds = [(0.35, 0.65)]
        opt_tol = 1e-8
        res = optimize.minimize(self._obj_func, x0 = 0.5, bounds=bnds, 
                                args=(amt_tkn_in, lp, tkn, lwr_tick, upr_tick), 
                                method='Nelder-Mead', tol=opt_tol)
        return res.x[0]
    
    def _obj_func(self, alpha, amt_tkn_in, lp, token_in, lwr_tick, upr_tick):
        opt_tol = 1e-8         
        swap_in = amt_tkn_in*alpha
        amt_tkn0, sqrtp_cur  = UniV3Helper().quote(lp, token_in, swap_in, lwr_tick, upr_tick)
        sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/2**96
        sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/2**96    
        
        if(token_in.token_name == lp.token0):
            sqrtp_cur = 1/sqrtp_cur
            
        dPy = (sqrtp_cur - sqrtp_pa)
        dPx = (1/sqrtp_cur - 1/sqrtp_pb) 
        
        if(token_in.token_name == lp.token0):
            dL = amt_tkn0/dPy 
            amt_deposit = dL*dPx
        elif(token_in.token_name == lp.token1): 
            dL = amt_tkn0/dPx 
            amt_deposit = dL*dPy
        
        diff = amt_tkn_in - (swap_in + amt_deposit) 
        return abs(diff)+opt_tol
    
    def _calc_univ2_deposit_portion(self, lp, token_in, dx):

        tkn_supply = self._get_tkn_supply(lp, token_in)
        a = 997*(dx**2)/(1000*tkn_supply)
        b = dx*(1997/1000)
        c = -dx

        alpha = -(b - math.sqrt(b*b - 4*a*c)) / (2*a)
        return alpha 

    def _get_tkn_supply(self, lp, token_in):
        tokens = lp.factory.token_from_exchange[lp.name]
        if(lp.version == UniswapExchangeData.VERSION_V2):
            if(token_in.token_name == lp.token0):
                tkn_supply = lp.get_reserve(tokens[lp.token0])
            else:    
                tkn_supply = lp.get_reserve(tokens[lp.token1])
        elif(lp.version == UniswapExchangeData.VERSION_V3):   
            if(token_in.token_name == lp.token0):
                tkn_supply = lp.get_reserve(tokens[lp.token0])
            else:    
                tkn_supply = lp.get_reserve(tokens[lp.token1])
        return tkn_supply        
        




# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass

@dataclass
class ExchangeData:
    symbol: str
    address : str    


from .SolveDeltas import SolveDeltas
from .SimpleLPSimulation import SimpleLPSimulation
from .MarkovState import MarkovState
from .Arbitrage import Arbitrage
from .CorrectReserves import CorrectReserves
from .QuantTerminal import QuantTerminal



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..simulate import SolveDeltas
from ..process.deposit import SwapDeposit
from ..process.swap import WithdrawSwap
from ..utils.data import UniswapExchangeData
import numpy as np

X0 = 1
FAC = 1
MAX_ATTEMPTS = 5
USER_NM = 'reserve_correction'

class CorrectReserves:
    
    """ 
        Applies SolveDeltas to Correct x/y reserve amounts so that price reflects desired input price; 
        in the marjority of cases, the input price would be the most recent outside market price 
        
        Parameters
        -----------------
        lp : UniswapExchange
            Uniswap Exchange class
        x0 : float
            Initial market price at the beginning of the simulation 
        fac : float
            scipy.fsolve parameter for non-linear solver found in SolveDeltas class
    """             
    
    def __init__(self, lp, x0 = None, fac = None):
        self.lp = lp
        self.sDel = SolveDeltas(lp)
        self.x0 = X0 if x0 == None else int(x0)
        self.fac = FAC if fac == None else fac
        self.swap_dx = 0
        self.swap_dy = 0
        self.user_nm = USER_NM

    def set_user(self, user_nm): 
        self.user_nm = user_nm    
    
    def get_swap_dx(self): 
        
        """ get_swap_dx

            Get delta x reserve adjustment from SolveDeltas class
                
            Returns
            -----------------
            swap_dx : float
                Delta x reserve adjustment from SolveDeltas class                   
        """         
        
        return self.swap_dx
    
    def get_swap_dy(self): 
        
        """ get_swap_dy

            Get delta y reserve adjustment from SolveDeltas class
                
            Returns
            -----------------
            swap_dy : float
                Delta y reserve adjustment from SolveDeltas class                   
        """           
        
        return self.swap_dy    
    
    def apply(self, p, lwr_tick = None, upr_tick = None): 
        
        """ apply

            Apply reserve correction of mock pool given external price (eg. market price from 0x API)
                
            Parameters
            -----------------
            p : float
                Price (eg. market price from 0x API)               
        """         
        
        do_update = True  
        max_tries = 5; c = 0
        while(do_update and c <= MAX_ATTEMPTS):
            c=+1
            try: 
                self._update(p, lwr_tick, upr_tick)
                do_update = False 
            except:
                p = p + np.random.normal(0, 0.001)  
                do_update = True 
                
        assert c != MAX_ATTEMPTS, 'Uniswap: RESERVE_CORRECTION_FAILURE'   
        
    def get_x_tkn(self): 
        
        """ get_x_tkn

            Get x token from Uniswap mock pool
                
            Returns
            -----------------
            x_tkn : ERC20
                x token from mock pool                   
        """         
        
        return self.lp.factory.token_from_exchange[self.lp.name][self.lp.token0]
        
    def get_y_tkn(self):     
        
        """ get_y_tkn

            Get y token from Uniswap mock pool
                
            Returns
            -----------------
            y_tkn : ERC20
                y token from mock pool                   
        """           
        
        return self.lp.factory.token_from_exchange[self.lp.name][self.lp.token1]        
                
    def _update(self, p, lwr_tick, upr_tick):  
        tkn_x = self.get_x_tkn()
        tkn_y = self.get_y_tkn()
        self.swap_dx, self.swap_dy = self.sDel.calc(p, self.x0, self.fac)

        if(self.lp.version == UniswapExchangeData.VERSION_V2):
            if(self.swap_dx >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_x, self.user_nm, abs(self.swap_dx))
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_y, self.user_nm, abs(self.swap_dy))
            elif(self.swap_dy >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_y, self.user_nm, abs(self.swap_dy))
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_x, self.user_nm, abs(self.swap_dx)) 
               
        elif(self.lp.version == UniswapExchangeData.VERSION_V3):
            if(self.swap_dx >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_x, self.user_nm, abs(self.swap_dx), lwr_tick, upr_tick)
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_y, self.user_nm, abs(self.swap_dy), lwr_tick, upr_tick)
            elif(self.swap_dy >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_y, self.user_nm, abs(self.swap_dy), lwr_tick, upr_tick)
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_x, self.user_nm, abs(self.swap_dx), lwr_tick, upr_tick)


    def get_ticks(self, tkn_x):
        lwr_tick = UniV3Helper().get_tick_price(self.lp, -1, lp.get_price(tkn_x), 1000)
        upr_tick = UniV3Helper().get_tick_price(self.lp, 1, lp.get_price(tkn_x), 1000)
        return lwr_tick, upr_tick
              
         


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

class Wallets():
    
    def __init__(self):
        self.accounts = {}
        self.tokens = {}
        
    def update(self, _to, tkn, amt):
        
        if(amt >= 0):
            self.deposit(_to, tkn, amt)
        else:
            self.remove(_to, tkn, abs(amt))
        
    def deposit(self, _to, tkn, amt):
             
        if self.__chk_tkn(tkn):
            self.tokens[tkn.token_name]['total_amount'] += amt
            self.tokens[tkn.token_name]['last_deposit'] = amt
        else:
            self.__init_tkn(tkn, amt)
            
        if self.__chk_nm(_to):
            if self.__chk_tkn_nm(_to, tkn):
                self.accounts[_to][tkn.token_name]['amount'] += amt
            else:
                self.accounts[_to][tkn.token_name] = {}
                self.accounts[_to][tkn.token_name]['amount'] = amt                
        else:
            self.accounts[_to] = {}
            self.accounts[_to][tkn.token_name] = {}
            self.accounts[_to][tkn.token_name]['amount'] = amt
            

    def remove(self, _to, tkn, amt):   
              
        removed = True
        
        if self.__chk_tkn_nm(_to, tkn):
            if(amt <= self.accounts[_to][tkn.token_name]['amount']):
                self.accounts[_to][tkn.token_name]['amount'] -= amt
                self.tokens[tkn.token_name]['total_amount'] -= amt
                self.tokens[tkn.token_name]['last_deposit'] = -amt
            else:
                print('WALLET ERROR: not enough amount in {} account'.format(_to))
                removed = False
        else:
            print('WALLET ERROR: token not in vault')
            removed = False
            
        return removed  
    
    def get_token_amount(self, tkn):
        return self.tokens[tkn.token_name]['total_amount']
    
    def __init_tkn(self, token, amt):        
        self.tokens[token.token_name] = {}
        self.tokens[token.token_name]['token'] = token 
        self.tokens[token.token_name]['total_amount'] = amt
        self.tokens[token.token_name]['last_deposit'] = amt 
        
    def __chk_tkn(self, token):        
        return True if token.token_name in self.tokens else False  
    
    def __chk_tkn_nm(self, _to, token):        
        if self.__chk_nm(_to): 
            return True if token.token_name in self.accounts[_to] else False
        else:
            return False    
        
    def __chk_nm(self, _to):        
        return True if _to in self.accounts else False        


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Modified version of MIT licenced original utils file from chainflip-io 
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

import sys, traceback, math, copy
from decimal import *
import pytest

from .Shared import *

TEST_TOKEN0 = "Token0"
TEST_TOKEN1 = "Token1"

@dataclass
class UniV3Utils():

    TEST_TOKENS = [TEST_TOKEN0, TEST_TOKEN1]
    
    @dataclass
    class FeeAmount:
        LOW: int = 500
        MEDIUM: int = 3000
        HIGH: int = 10000
    
    def getMinTick(tickSpacing):
        return math.ceil(MIN_TICK / tickSpacing) * tickSpacing
    
    
    def getMaxTick(tickSpacing):
        return math.floor(MAX_TICK / tickSpacing) * tickSpacing
    
    
    def getMaxLiquidityPerTick(tickSpacing):
        denominator = (getMaxTick(tickSpacing) - getMinTick(tickSpacing)) // tickSpacing + 1
        return (2**128 - 1) // denominator
    
    TICK_SPACINGS = {FeeAmount.LOW: 10, FeeAmount.MEDIUM: 60, FeeAmount.HIGH: 200}   
    
    def encodePriceSqrt(reserve1, reserve0):
        # Workaround to get the same numbers as JS
        # This ratio doesn't output the same number as in JS using big number. This causes some
        # disparities in the expected results. Full ratios (1,1), (2,1) ...
        # Forcing values obtained by bigNumber.js when ratio is not exact.
        if reserve1 == 121 and reserve0 == 100:
            return 87150978765690771352898345369
        elif reserve1 == 101 and reserve0 == 100:
            return 79623317895830914510487008059
        elif reserve1 == 1 and reserve0 == 10:
            return 25054144837504793118650146401
        elif reserve1 == 1 and reserve0 == 2**127:
            return 6085630636
        elif reserve1 == 2**127 and reserve0 == 1:
            return 1033437718471923701407239276819587054334136928048
        else:
            return int(math.sqrt(reserve1 / reserve0) * 2**96)
    
    
    def expandTo18Decimals(number):
        # Converting to int because python cannot shl on a float
        return int(number * 10**18)
    
    
    # @dev This function will handle reverts (aka assert failures) in the tests. However, in python there is no revert
    # as in the blockchain. So we will create a hard copy of the current pool and call the same method there.
    def tryExceptHandler(fcn, assertMessage, *args):
    
        reverted = False
    
        try:
            # reference to object
            pool = fcn.__self__
            fcnName = fcn.__name__
    
            # hard copy to prevent state changes in the pool
            poolCopy = copy.deepcopy(pool)
    
            try:
                fcn = getattr(poolCopy, fcnName)
            except AttributeError:
                assert "Function not found in pool: " + fcnName
        except:
            # e.g. case when swapExact1ForZero is expected to revert
            print(
                "Non-pool class function passed - expect pool to be copied as part of the call"
            )
    
        try:
            fcn(*args)
        except AssertionError as msg:
            reverted = True
            _, _, tb = sys.exc_info()
            traceback.print_tb(tb)  # Fixed format
            tb_info = traceback.extract_tb(tb)
            filename, line, func, text = tb_info[-1]
    
            print("An error occurred on line {} in statement {}".format(line, text))
            if str(msg) != assertMessage:
                print(
                    "Reverted succesfully but not for the expected reason. \n Expected: '"
                    + str(assertMessage)
                    + "' but got: '"
                    + str(msg)
                    + "'"
                )
                assert False
            print("Succesful revert")
    
        if not reverted:
            print("Failed to revert: " + assertMessage)
            assert False
    
    
    def getPositionKey(address, lowerTick, upperTick):
        return hash((address, lowerTick, upperTick))
    
    
    def getLimitPositionKey(address, tick, isToken0):
        return hash((address, tick, isToken0))
    
    
    ### POOL SWAPS ###
    def swapExact0For1(pool, amount, recipient, sqrtPriceLimit):
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else getSqrtPriceLimitX96(TEST_TOKEN0)
        )
        return swap(pool, TEST_TOKEN0, [amount, 0], recipient, sqrtPriceLimitX96)
    
    
    def swap0ForExact1(pool, amount, recipient, sqrtPriceLimit):
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else getSqrtPriceLimitX96(TEST_TOKEN0)
        )
        return swap(pool, TEST_TOKEN0, [0, amount], recipient, sqrtPriceLimitX96)
    
    
    def swapExact1For0(pool, amount, recipient, sqrtPriceLimit):
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else getSqrtPriceLimitX96(TEST_TOKEN1)
        )
        return swap(pool, TEST_TOKEN1, [amount, 0], recipient, sqrtPriceLimitX96)
    
    
    def swap1ForExact0(pool, amount, recipient, sqrtPriceLimit):
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else getSqrtPriceLimitX96(TEST_TOKEN1)
        )
        return swap(pool, TEST_TOKEN1, [0, amount], recipient, sqrtPriceLimitX96)
    
    
    def swapToLowerPrice(pool, recipient, sqrtPriceLimit):
        return pool.swap(recipient, True, MAX_INT256, sqrtPriceLimit)
    
    
    def swapToHigherPrice(pool, recipient, sqrtPriceLimit):
        return pool.swap(recipient, False, MAX_INT256, sqrtPriceLimit)
    
    
    def swap(self, pool, inputToken, amounts, recipient, sqrtPriceLimitX96):
        [amountIn, amountOut] = amounts
        exactInput = amountOut == 0
        amount = amountIn if exactInput else amountOut
    
        if inputToken == TEST_TOKEN0:
            if exactInput:
                checkInt128(amount)
                return pool.swap(recipient, True, amount, sqrtPriceLimitX96)
            else:
                checkInt128(-amount)
                return pool.swap(recipient, True, -amount, sqrtPriceLimitX96)
        else:
            if exactInput:
                checkInt128(amount)
                return pool.swap(recipient, False, amount, sqrtPriceLimitX96)
            else:
                checkInt128(-amount)
                return pool.swap(recipient, False, -amount, sqrtPriceLimitX96)
    
    
    def getSqrtPriceLimitX96(inputToken):
        if inputToken == TEST_TOKEN0:
            return MIN_SQRT_RATIO + 1
        else:
            return MAX_SQRT_RATIO - 1
    
    
    ################
    
    
    def formatPrice(price):
        fraction = (price / (2**96)) ** 2
        return formatAsInSnapshot(fraction)
    
    
    def formatTokenAmount(amount):
        fraction = amount / (10**18)
        return formatAsInSnapshot(fraction)
    
    
    def formatAsInSnapshot(number):
        # To match snapshot formatting
        precision = int(f"{number:e}".split("e")[-1])
        # For token we want 4 extra decimals of precision
        if precision >= 0:
            precision = 4
        else:
            precision = -precision + 4
    
        return format(number, "." + str(precision) + "f")
    
    
    def formatPriceWithPrecision(price, precision):
        fraction = (price / (2**96)) ** 2
        return round(fraction, precision)



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass
from .ExchangeData import ExchangeData
from ...erc import ERC20

DEFAULT_VERSION = 'V2'
DEFAULT_TYPE = 'DEC'

@dataclass
class UniswapExchangeData(ExchangeData):

    VERSION_V2 = DEFAULT_VERSION
    VERSION_V3 = 'V3'

    TYPE_DEC = DEFAULT_TYPE
    TYPE_GWEI = 'GWEI'    
        
    tkn0: ERC20
    tkn1: ERC20 
    version: str = DEFAULT_VERSION
    precision: str = DEFAULT_TYPE
    tick_spacing: int = None   
    fee: int = None


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ..deposit import SwapDeposit
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel

class SwapIndexMint(Process):
    
    """ Process to swap-deposit single token into LP and mint its respective indexing token

        Parameters
        ----------
        ivault : IndexVault
            Index vault     
        opposing : boolean
            Boolean variable to determine whether to mint token opposing the one that was deposited                
    """     

    def __init__(self, ivault, opposing = None):
        self.ivault = IndexVault('iVault', "0x7") if ivault  == None else ivault
        self.opposing = False if opposing  == None else opposing
        self.mint_amt = 0
            
    def apply(self, lp_tkn, token, user_nm, amt, lwr_tick = None, upr_tick = None):    
        
        """ apply

            Swap-deposit single token into LP and mint its respective indexing token
                
            Parameters
            -------
            lp_tkn : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amt : float
               token amount to be swapped and minted 
                
            Returns
            -------
            (amount_in, amount_out) : float, float
                token swap amounts                
        """           
        
        lp_amt0 = lp_tkn.total_supply
        out = SwapDeposit().apply(lp_tkn, token, user_nm, amt, lwr_tick, upr_tick)
        lp_amt1 = lp_tkn.total_supply
        lp_amt = lp_tkn.last_liquidity_deposit

        mint_token = token if not self.opposing else self.get_trading_token(lp_tkn, token)
        self.ivault.deposit_lp_tkn(user_nm, lp_tkn, lp_amt)
        self.ivault.rebase_index_tkn(lp_tkn, mint_token, lwr_tick, upr_tick) 
     
    def get_trading_token(self, lp, token):
        
        """ get_trading_token

            Get opposing token from specified token
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token      
                
            Returns
            -------
            trading_token : ERC20 
                opposing ERC20 token                   
        """           
        
        tokens = lp.factory.exchange_to_tokens[lp.name]
        trading_token = tokens[lp.token1] if token.token_name == lp.token0 else tokens[lp.token0]
        return trading_token           
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...erc import ERC20
from .Vault import Vault 
from ..index import RebaseIndexToken

INDEX_TOKEN = 'INDEX'

class IndexVault(Vault):
    
    """ Vault of index tokens

        Parameters
        ----------
        token_name : str
            Token name 
        token_addr : str
            Token address  
        lp_providers : dictionary
            Map of LP providers to their respective holdings
        lp_tokens : dictionary
            Map of LPs to their ERC20 tokens and total amounts
        index_tokens : dictionary
            Map of index tokens to their ERC20 tokens and total amounts            
    """  
    
    
    def __init__(self, name: str, addr: str) -> None:
        self.token_name = name
        self.token_addr = addr 
        self.lp_providers = {}
        self.lp_tokens = {}
        self.index_tokens = {}
      
    def rebase_index_tkn(self, lp_token, token = None, lwr_tick = None, upr_tick = None):
        
        """ rebase_index_tkn

            Mint index token from LP and specified token from pair
                
            Parameters
            -------
            lp_token : DAOSYSERC20
                LP token 
            token : ERC20
                specified token from pair                                 
        """         
        
        if(token != None):
            self.__rebase_index_tkn(lp_token, token, lwr_tick, upr_tick)
            self.update_accounts(lp_token, token, lwr_tick, upr_tick)
        else:  
            tokens = lp_token.factory.exchange_to_tokens[lp_token.name]
            self.__rebase_index_tkn(lp_token, tokens[lp_token.token0], lwr_tick, upr_tick)
            self.__rebase_index_tkn(lp_token, tokens[lp_token.token1], lwr_tick, upr_tick)
            self.update_accounts(lp_token, tokens[lp_token.token0])
            self.update_accounts(lp_token, tokens[lp_token.token1])  
            self.update_index_tkn(lp_token, token)    
        
    def deposit_lp_tkn(self, _to, lp_tkn, amt):
        
        """ deposit_lp_tkn

            Deposit specified amount of LP token to a user account
                
            Parameters
            -------
            _to : str
                account holder            
            token : DAOSYSERC20
                LP token 
            amt : float
                amount of LP to be deposited                                
        """          
        
        if self.__chk_tkn(lp_tkn):
            self.lp_tokens[lp_tkn.token_name]['total_amount'] += amt
            self.lp_tokens[lp_tkn.token_name]['last_deposit'] = amt
        else:
            self.__init_tkn(lp_tkn, amt)
            
        if self.__chk_nm(_to):
            if self.__chk_tkn_nm(_to, lp_tkn):
                self.lp_providers[_to][lp_tkn.token_name]['amount'] += amt
            else:
                self.lp_providers[_to][lp_tkn.token_name] = {}
                self.lp_providers[_to][lp_tkn.token_name]['amount'] = amt                
        else:
            self.lp_providers[_to] = {}
            self.lp_providers[_to][lp_tkn.token_name] = {}
            self.lp_providers[_to][lp_tkn.token_name]['amount'] = amt    
            

    def remove_lp_tkn(self, _to, lp_tkn, amt):   
        
        """ remove_lp_tkn

            Remove specified amount of LP token from a user account
                
            Parameters
            -------
            _to : str
                account holder            
            lp_tkn : DAOSYSERC20
                LP token 
            amt : float
                amount of LP to be deposited                                
        """       
        
        removed = True
        
        if self.__chk_tkn_nm(_to, lp_tkn):
            if(amt <= self.lp_providers[_to][lp_tkn.token_name]['amount']):
                self.lp_providers[_to][lp_tkn.token_name]['amount'] -= amt
                self.lp_tokens[lp_tkn.token_name]['total_amount'] -= amt
                self.lp_tokens[lp_tkn.token_name]['last_deposit'] = -amt
            else:
                print('ERROR: not enough amount in {} account'.format(_to))
                removed = False
        else:
            print('ERROR: token not in vault')
            removed = False
            
        return removed    
            
    def burn_lp_token(self, _from, lp_tkn, token, lp_amt):
        
        """ burn_lp_token

            Burn specified amount of LP token from a user account referencing its parent pairing token
                
            Parameters
            -------
            _from : str
                account holder            
            lp_tkn : DAOSYSERC20
                LP token
            token : ERC20
                parent pairing token                
            lp_amt : float
                amount of LP to be burned
                
            Returns
            -------
            res : dictionary
                map of pair tokens and their respective amounts                 
        """          
        
        res = {}
        if(self.check_token_account(token, _from)):
            removed = self.remove_lp_tkn(_from, lp_tkn, lp_amt)
            if(removed):
                amount0 = lp_amt*lp_tkn.reserve0/lp_tkn.total_supply
                amount1 = lp_amt*lp_tkn.reserve1/lp_tkn.total_supply              
                lp_tkn.remove_liquidity(_from, lp_amt, amount0, amount1)
                self.rebase_index_tkn(lp_tkn, token)
                self.update_accounts(lp_tkn, token)            
            
        return {lp_tkn.token0: amount0, lp_tkn.token1: amount1} if removed else res

            
    def get_tkn_pair_amount(self, lp_tkn, token, liq, lwr_tick = None, upr_tick = None):
        
        """ get_tkn_pair_amount

            Get amount of specified token from LP pairing 
                
            Parameters
            -------         
            lp_tkn : DAOSYSERC20
                LP token
            token : ERC20
                parent pairing token                
            liq : float
                amount of LP to be burned   
                
            Returns
            -------
            amt : float
                amount of specified token from pairing                 
        """          
        amt = RebaseIndexToken().apply(lp_tkn, token, liq, lwr_tick, upr_tick)
        return amt
    
    def get_token_type(self, token):
        
        """ get_token_type

            Get token type (ie, LP, ERC20, Index)
                
            Parameters
            -------         
            token : ERC20
                parent pairing token                
                
            Returns
            -------
            type : str
                token type                  
        """         
        
        return token.__class__.__name__   
    
    def update_accounts(self, lp_tkn, tkn, lwr_tick, upr_tick): 

        """ update_accounts

            Update specified token across all accounts
                
            Parameters
            -------         
            lp_tkn : DAOSYSERC20
                parent pairing token                
            tkn : ERC20
                parent pairing token

            Returns
            -------
            type : str
                token type                  
        """         
        
        index_tokens = {}
        exchange = lp_tkn.token_name
        index_tokens[tkn.token_name] = 'i'+tkn.token_name
        
        self.update_index_tkn(lp_tkn, tkn, lwr_tick, upr_tick)
        
        new_total = 0
        tkn_nm = tkn.token_name
        for account in self.lp_providers:
            amt = self.lp_providers[account][exchange]['amount']
            exch_tkn = lp_tkn.factory.token_from_exchange[lp_tkn.name][tkn_nm]
            self.lp_providers[account][index_tokens[tkn_nm]] = {}
            self.lp_providers[account][index_tokens[tkn_nm]]['amount'] = self.get_tkn_pair_amount(lp_tkn, exch_tkn, amt, lwr_tick, upr_tick)             
                  
    def update_account(self, lp_tkn, tkn, _from): 
        
        """ update_accounts

            Update specified token from specified account
                
            Parameters
            -------         
            lp_tkn : DAOSYSERC20
                parent pairing token                
            tkn : ERC20
                parent pairing token
            _from : str
                account holder                                
        """         

        index_tokens = {}
        exchange = lp_tkn.token_name
        index_tokens[tkn.token_name] = 'i'+tkn.token_name

        for tkn_nm in index_tokens.keys():
            amt = self.lp_providers[_from][exchange]['amount']
            exch_tkn = lp_tkn.factory.token_from_exchange[lp_tkn.name][tkn_nm]
            self.lp_providers[_from][index_tokens[tkn_nm]] = {}
            self.lp_providers[_from][index_tokens[tkn_nm]]['amount'] = self.get_tkn_pair_amount(lp_tkn, exch_tkn, amt)  
            
        if(self.lp_providers[_from][lp_tkn.token_name]['amount'] == 0):
            self.lp_providers.pop(_from)            
                
    def get_account(self, _from):
        
        """ get_account

            Get LP information from account holder
                
            Parameters
            -------         
            _from : str
                account holder                

            Returns
            -------
            type : dictionary
                LP account with amount of holdings                
        """           
        
        return self.lp_providers[_from]
    
    
    def check_token_account(self, token, _from):
        
        """ check_token_account

            Check if account holder is holding specified token
                
            Parameters
            -------  
            token : ERC20
                specified query token             
            _from : str
                account holder                

            Returns
            -------
            is_tkn_account : boolean
                boolean if account exists for specified token                
        """             
        
        is_tkn_account = False
        for tkn_nm in self.lp_providers[_from].keys():
            is_tkn_account = True if tkn_nm == 'i'+token.token_name and not is_tkn_account else is_tkn_account
        return is_tkn_account    
    
    def __rebase_index_tkn(self, lp_token, token, lwr_tick, upr_tick):
        
        """ __rebase_index_tkn

            Mint index token based on the amount of LP in account
                
            Parameters
            -------  
            lp_token : ERC20
                specified parent token             
            token : DAOSYSERC20
                LP token                              
        """          
        
        mint_tkn_name = 'i'+token.token_name
        liq = self.lp_tokens[lp_token.token_name]['total_amount']
        last_liq = self.lp_tokens[lp_token.token_name]['last_deposit']
        delta_amt = self.get_tkn_pair_amount(lp_token, token, last_liq, lwr_tick, upr_tick)
        amt = self.get_tkn_pair_amount(lp_token, token, liq, lwr_tick, upr_tick)
        tkn = ERC20(mint_tkn_name, None)
        tkn.deposit(None, amt) 
        
        if mint_tkn_name not in self.index_tokens:    
            self.index_tokens[mint_tkn_name] = {}
            self.index_tokens[mint_tkn_name]['token'] = tkn 
            self.index_tokens[mint_tkn_name]['total'] = self.get_tkn_pair_amount(lp_token, token, liq, lwr_tick, upr_tick) 
            self.index_tokens[mint_tkn_name]['total_lp'] = liq
            self.index_tokens[mint_tkn_name]['last_deposit'] = tkn.token_total
            self.index_tokens[mint_tkn_name]['last_lp_deposit'] = amt
            self.index_tokens[mint_tkn_name]['parent_token'] = token.token_name
            self.index_tokens[mint_tkn_name]['exchange'] = lp_token.token_name
        else:          
            prev_total = self.index_tokens[mint_tkn_name]['total']
            self.index_tokens[mint_tkn_name]['token'] = tkn 
            self.index_tokens[mint_tkn_name]['total'] = self.get_tkn_pair_amount(lp_token, token, liq, lwr_tick, upr_tick) 
            self.index_tokens[mint_tkn_name]['total_lp'] = liq
            self.index_tokens[mint_tkn_name]['last_deposit'] = delta_amt
            self.index_tokens[mint_tkn_name]['last_lp_deposit'] = last_liq
               
            
    def update_index_tkn(self, lp_token, token, lwr_tick, upr_tick):   
        mint_tkn_name = 'i'+token.token_name
        if mint_tkn_name in self.index_tokens:  
            liq = self.lp_tokens[lp_token.token_name]['total_amount']
            last_liq = self.lp_tokens[lp_token.token_name]['last_deposit']
            self.index_tokens[mint_tkn_name]['total'] = self.get_tkn_pair_amount(lp_token, token, liq, lwr_tick, upr_tick)  
            self.index_tokens[mint_tkn_name]['last_deposit'] = self.get_tkn_pair_amount(lp_token, token, last_liq, lwr_tick, upr_tick)  
            self.index_tokens[mint_tkn_name]['total_lp'] = liq  
            self.index_tokens[mint_tkn_name]['last_lp_deposit'] = last_liq             
            
    def __chk_nm(self, _to):
        
        """ __chk_nm

            Determine if account exisits under holder name 
                
            Parameters
            -------  
            _to : str
                account holder to check 
                
            Returns
            -------
            is_account : boolean
                boolean of whether account exisits                   
        """          
        
        return True if _to in self.lp_providers else False

    def __init_tkn(self, token, amt):
        
        """ __init_tkn

            Initialize token at specified amount 
                
            Parameters
            -------  
            token : ERC20
                specified token   
            amt : float
                amount of new inititalized token                 
        """          
        
        self.lp_tokens[token.token_name] = {}
        self.lp_tokens[token.token_name]['token'] = token 
        self.lp_tokens[token.token_name]['total_amount'] = amt
        self.lp_tokens[token.token_name]['last_deposit'] = amt
     
    def __chk_tkn(self, lp_token):
        
        """ __chk_tkn

            Determine if LP token exisits
                
            Parameters
            -------  
            lp_token : DAOSYSERC20
                LP token to check 
                
            Returns
            -------
            is_token : boolean
                boolean of whether LP token exisits                   
        """         
        
        return True if lp_token.token_name in self.lp_tokens else False

    def __chk_tkn_nm(self, _to, lp_token):
        
        """ __chk_tkn_nm

            Determine if LP token exisits for specified account 
                
            Parameters
            -------  
            _to : str
                account name             
            lp_token : DAOSYSERC20
                LP token to check 
                
            Returns
            -------
            is_chk_nm : boolean
                boolean of whether LP token exisits for specified account                  
        """          
        
        if self.__chk_nm(_to): 
            return True if lp_token.token_name in self.lp_providers[_to] else False
        else:
            return False            
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData
import math

class Swap(Process):
    
    """ Process to swap token X for token Y (and vice verse) 

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts                 
    """       

    def __init__(self, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
            
    def apply(self, lp, token_in, user_nm, amount_in, sqrt_price_limit = None):    
        
        """ apply

            Swap token X for token Y (and vice verse) 
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_in : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_in : float
                token amount to be swap 
            lwr_tick : int
                lower tick of the position in which to add liquidity   
            upr_tick : int
                upper tick of the position in which to add liquidity                
                
            Returns
            -------
            amount_out_expected : float
                exchanged token amount               
        """          
        
        amount_in = tDel.delta() if amount_in == None else amount_in
        
        if(lp.version == UniswapExchangeData.VERSION_V2):
            amount_out = math.floor(lp.get_amount_out(amount_in, token_in))
            amount_out_expected = lp.swap_exact_tokens_for_tokens(amount_in, amount_out, token_in, to_addr=user_nm)
            
        elif(lp.version == UniswapExchangeData.VERSION_V3):
            tokens = lp.factory.token_from_exchange[lp.name]
            if(token_in.token_name == lp.token0):
                amount_out = lp.swapExact0For1(user_nm, amount_in, sqrt_price_limit)
                amount_out_expected = abs(amount_out[2])
            else: 
                amount_out = lp.swapExact1For0(user_nm, amount_in, sqrt_price_limit)
                amount_out_expected = abs(amount_out[1])
            
        return amount_out_expected        
        



from .UniV3Utils import UniV3Utils
from .UniV3Helper import UniV3Helper
#from .LiquidityMath import LiquidityMath
#from .Position import Position
#from .Shared import Shared
#from .SqrtPriceMath import SqrtPriceMath
#from .SwapMath import SwapMath
#from .Tick import Tick
#from .TickMath import TickMath


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from .LPQuote import LPQuote

class LPTokenQuote():
    
    def __init__(self, lwr_tick = None, upr_tick = None, quote_child = True):
        self.quote_child = quote_child 
        self.lwr_tick = lwr_tick
        self.upr_tick = upr_tick         
 
    def get_x(self, lp, amt_tkn_x): 
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0] 
        parent_lp = self.get_base_lp(lp, tkn_x)
        parent_lp_x_tkn = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token0]
        parent_x_tkn = tkn_x.parent_tkn if tkn_x.type == 'index' else tkn_x

        if(parent_x_tkn.token_name != parent_lp_x_tkn.token_name):
            amt_tkn_x = LPQuote().get_amount(parent_lp, parent_lp_x_tkn, amt_tkn_x, self.lwr_tick, self.upr_tick)     

        if(tkn_x.type == 'standard'):
            amt_x = LPQuote(False).get_amount(parent_lp, parent_x_tkn, amt_tkn_x, self.lwr_tick, self.upr_tick)
            lp_amt = LPQuote().get_lp_from_amount(lp, parent_x_tkn, amt_x, self.lwr_tick, self.upr_tick)
        else: 
            amt_x = LPQuote(False).get_amount(parent_lp, parent_x_tkn, amt_tkn_x, self.lwr_tick, self.upr_tick) 
            amt_x_lp = LPQuote().get_lp_from_amount(parent_lp, parent_x_tkn, amt_x, self.lwr_tick, self.upr_tick)
            lp_amt = LPQuote().get_lp_from_amount(lp, parent_x_tkn, amt_x_lp, self.lwr_tick, self.upr_tick)

        return lp_amt 
        
    def base_x_asset(self, lp):
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        if(tkn_x.type == 'index'):
            tkn = tkn_x.parent_tkn
        else:
            tkn = tkn_x

        return tkn    

    def get_base_lp(self, lp, tkn):
        if(tkn.type == 'index'):
            base_lp = tkn.parent_lp
        else:
            base_lp = lp.factory.parent_lp
        return base_lp if base_lp != None else lp      
 
 

        
            



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from abc import *

class Process(ABC):
             
    @abstractmethod        
    def apply(self, lp, token, user_nm, amount):
        pass



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Utils
from ...utils.tools.v3 import UniV3Helper
import math

class Join():
    
    """ Process to join x and y amounts to pool              
    """       

    def __init__(self):
        pass

    def apply(self, lp, user_nm, amount0, amount1, lwr_tick = None, upr_tick = None):
        """ apply

            Join x and y amounts to pool
                
            Parameters
            -------
            lp : Exchange
                LP exchange            
            user_nm : str
                account name
            amount0 : float
               x token amount 
            amount1 : float
               y token amount       
            lwr_tick : int
               Lower tick of the position in which to add liquidity   
            upr_tick : int
               Upper tick of the position in which to add liquidity   
                     
            Returns
            -------
            out : dictionary
                join output               
        """ 

        if(lp.version == UniswapExchangeData.VERSION_V2):
            out = lp.add_liquidity(user_nm, amount0, amount1, amount0, amount1)
            
        elif(lp.version == UniswapExchangeData.VERSION_V3):
            init_price = UniV3Utils.encodePriceSqrt(amount1, amount0)
            sqrtP = init_price/2**96
            Ly = UniV3Helper().calc_Ly(sqrtP, amount1, lwr_tick, upr_tick)
            Lx = UniV3Helper().calc_Lx(sqrtP, amount0, lwr_tick, upr_tick)
            L_mint = min(Ly, Lx)    
            lp.initialize(init_price)
            out = lp.mint(user_nm, lwr_tick, upr_tick, L_mint)
             
        return out  


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Helper

class AddLiquidity(Process):
     
    """ Add liquidity process

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts        
    """     

    def __init__(self, init_price = None, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
        self.init_price = 1 if init_price == None else init_price
            
    def apply(self, lp, token_in, user_nm, amount_in, lwr_tick = None, upr_tick = None):    
        
        """ apply

            Adds liquidity using only X or Y amounts
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_in : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_in : float
                token amount to be add to liquidity 
            lwr_tick : int
                lower tick of the position in which to add liquidity   
            upr_tick : int
                upper tick of the position in which to add liquidity                  
                
            Returns
            -------
            (amount_in, amount_out) : float, float
                token swap amounts                
        """          
        
        amount_in = tDel.delta() if amount_in == None else amount_in
        tokens = lp.factory.token_from_exchange[lp.name]  

        if(lp.version == UniswapExchangeData.VERSION_V2):

            res0 = lp.get_reserve(tokens[lp.token0])
            res1 = lp.get_reserve(tokens[lp.token1])
            tot_liq = lp.get_liquidity()
            
            if(token_in.token_name == lp.token0):
                balance0 = amount_in         
                if(res1 > 0):
                    balance1 = lp.quote(amount_in, res0, res1)
                else:
                    balance1 = self.init_price*balance0  
                    
                lp.add_liquidity(user_nm, balance0, balance1, balance0, balance1)
                
            elif(token_in.token_name == lp.token1):
                balance1 = amount_in
                if(res0 > 0):
                    balance0 = lp.quote(amount_in, res1, res0)
                else:
                    balance0 = self.init_price*balance1            
                
                lp.add_liquidity(user_nm, balance0, balance1, balance0, balance1) 

        elif(lp.version == UniswapExchangeData.VERSION_V3): 

            sqrt_P = lp.slot0.sqrtPriceX96/2**96
            if(token_in.token_name == lp.token0):
                Lx = UniV3Helper().calc_Lx(sqrt_P, amount_in, lwr_tick, upr_tick)
                (balance0, balance1)  = lp.mint(user_nm, lwr_tick, upr_tick, Lx)
            elif(token_in.token_name == lp.token1): 
                Ly = UniV3Helper().calc_Ly(sqrt_P, amount_in, lwr_tick, upr_tick)
                (balance0, balance1) = lp.mint(user_nm, lwr_tick, upr_tick, Ly)  

        return {lp.token0:balance0, lp.token1:balance1}    
    
 
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..index import RebaseIndexToken
from .LPQuote import LPQuote
import math

class TreeAmountQuote():
    
    def __init__(self, lwr_tick = None, upr_tick = None, quote_native_token = True, exchg_price = 1):
        self.exchg_price = exchg_price if exchg_price != 1 else exchg_price
        self.quote_native_token = quote_native_token 
        self.lwr_tick = lwr_tick
        self.upr_tick = upr_tick

    def get_tot_x(self, lp, amt0, amt1):
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]
        amt_in_x =  self.get_x(lp, tkn_x, amt0) + self.get_x(lp, tkn_y, amt1)
        return self.get_native_x(lp, tkn_x, amt_in_x) if self.quote_native_token else amt_in_x

    def get_tot_y(self, lp, amt0, amt1):
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]
        amt_in_y = self.get_y(lp, tkn_x, amt0) + self.get_y(lp, tkn_y, amt1)
        return self.get_native_y(lp, tkn_y, amt_in_y) if self.quote_native_token else amt_in_y 

    def get_x(self, lp, tkn, amt):
        amt_in_x = 0 
        if(tkn.type == 'standard'):
            x_tkn = self.base_x_asset_nm(lp) == tkn.token_name
            base_lp = self.get_base_lp(lp, tkn)
            amt_in_x = LPQuote(not x_tkn).get_amount(base_lp, tkn, amt)
        elif(tkn.type == 'index'):
            x_tkn = self.base_x_asset_nm(lp) == tkn.parent_tkn.token_name
            amt_in_x = LPQuote(not x_tkn).get_amount_from_lp(tkn.parent_lp, tkn.parent_tkn, amt, self.lwr_tick, self.upr_tick)

        return amt_in_x

    def get_y(self, lp, tkn, amt): 
        amt_in_y = 0 
        if(tkn.type == 'standard'):
            y_tkn = self.base_y_asset_nm(lp) == tkn.token_name
            base_lp = self.get_base_lp(lp, tkn)
            amt_in_y = LPQuote(not y_tkn).get_amount(base_lp, tkn, amt, self.lwr_tick, self.upr_tick)
        elif(tkn.type == 'index'):
            y_tkn = self.base_y_asset_nm(lp) == tkn.parent_tkn.token_name
            amt_in_y = LPQuote(not y_tkn).get_amount_from_lp(tkn.parent_lp, tkn.parent_tkn, amt, self.lwr_tick, self.upr_tick)
            
        return amt_in_y  
    
    def get_native_x(self, lp, tkn_x, x_amt):
        parent_lp = self.get_base_lp(lp, tkn_x)
        parent_lp_x_tkn = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token0]
        parent_x_tkn = tkn_x.parent_tkn if tkn_x.type == 'index' else tkn_x

        if(parent_x_tkn.token_name != parent_lp_x_tkn.token_name):
            x_amt = LPQuote().get_amount(parent_lp, parent_x_tkn, x_amt, self.lwr_tick, self.upr_tick) 

        return x_amt

    def get_native_y(self, lp, tkn_y, y_amt):
        parent_lp = self.get_base_lp(lp, tkn_y)
        parent_lp_y_tkn = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token1]
        parent_y_tkn = tkn_y.parent_tkn if tkn_y.type == 'index' else tkn_y

        if(parent_y_tkn.token_name != parent_lp_y_tkn.token_name):
            y_amt = LPQuote().get_amount(parent_lp, parent_y_tkn, y_amt, self.lwr_tick, self.upr_tick) 

        return y_amt    
    
    def base_x_asset_nm(self, lp):
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        if(tkn_x.type == 'index'):
            tkn_name = tkn_x.parent_tkn.token_name
        else:
            tkn_name = tkn_x.token_name

        return tkn_name
    
    def base_x_asset(self, lp):
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        if(tkn_x.type == 'index'):
            tkn = tkn_x.parent_tkn
        else:
            tkn = tkn_x

        return tkn    

    def base_y_asset_nm(self, lp):
        tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]
        if(tkn_y.type == 'index'):
            tkn_name = tkn_y.parent_tkn.token_name
        else:
            tkn_name = tkn_y.token_name

        return tkn_name

    def get_base_lp(self, lp, tkn):
        if(tkn.type == 'index'):
            base_lp = tkn.parent_lp
        else:
            base_lp = lp.factory.parent_lp
        return base_lp if base_lp != None else lp      



from .LPQuote import LPQuote
from .LPTokenQuote import LPTokenQuote
from .TreeAmountQuote import TreeAmountQuote
from .IndexTokenQuote import IndexTokenQuote
from .QuoteLiquidity import QuoteLiquidity



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ..liquidity import AddLiquidity
from ..mint import SwapIndexMint
from ...cpt.quote import LPQuote
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Helper

class JoinTree(Process):

    """ Process to join tree with liquidity 

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts        
    """     

    def __init__(self, init_price = None, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
        self.init_price = 1 if init_price == None else init_price
            
    def apply(self, child_lp, user_nm, iVault, parent_tkn_amt, lwr_tick = None, upr_tick = None):    
        
        """ apply

            Join x amounts to parent pool, and swap index mint into child LP
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            iVault : iVault
                specified index vault               
            user_nm : str
                account name
            lp_amt_in : float
                lp token amount to be added 
                                
            Returns
            -------
            (amount_in, amount_out) : float, float
                token swap amounts                
        """  

        lp_tkns = child_lp.factory.token_from_exchange[child_lp.name]
        x_tkn = lp_tkns[[*lp_tkns][0]]
        y_tkn = lp_tkns[[*lp_tkns][1]]
    
        parent_lp = child_lp.factory.parent_lp
        parent_lp_tkns = parent_lp.factory.token_from_exchange[parent_lp.name]
        parent_x_tkn = parent_lp_tkns[[*parent_lp_tkns][0]]
        parent_y_tkn = parent_lp_tkns[[*parent_lp_tkns][0]]    
        
        itkn_nm = y_tkn.token_name
        parent_tkn = parent_y_tkn
        parent_lp = child_lp.factory.parent_lp
        SwapIndexMint(iVault).apply(parent_lp, parent_tkn, user_nm, parent_tkn_amt)
        mint_itkn_deposit = iVault.index_tokens[itkn_nm]['last_lp_deposit']
        opposite_tkn = x_tkn.token_name != parent_tkn.token_name
        tkn_amount1 = LPQuote(opposite_tkn).get_amount_from_lp(parent_lp, parent_tkn, mint_itkn_deposit) 
        price_itkn = tkn_amount1/mint_itkn_deposit 
        return AddLiquidity(price_itkn).apply(child_lp, y_tkn, user_nm, mint_itkn_deposit) 
        



# IPS.py
# Author: Ian Moore ( utiliwire@gmail.com )
# Date: Sept 2022

from abc import *

DAYS_IN_YEAR = 365.25
HOURS_IN_DAY = 24
SECONDS_IN_HOUR = 3600

class IPS(ABC):

    SECONDS_YEAR = DAYS_IN_YEAR*HOURS_IN_DAY*SECONDS_IN_HOUR
    
    @abstractmethod
    def calc_ips_from_state(self, state):
        pass
    
    def calc_ips(self, A0, A1, t_delta):
        pass
        
    
    


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData


class RemoveLiquidity(Process):
    
    """ Remove liquidity process

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts        
    """       

    def __init__(self, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
            
    def apply(self, lp, token_in, user_nm, amount_in, lwr_tick = None, upr_tick = None):    
        
        """ apply

            Removes liquidity using only X or Y amounts
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_in : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_in : float
                token amount to be add to liquidity 
            lwr_tick : int
                lower tick of the position in which to add liquidity   
            upr_tick : int
                upper tick of the position in which to add liquidity                
                
            Returns
            -------
            reserve amounts : dictionary
                dictionary of reserve amounts                
        """          
        
        amount_in = self.tDel.delta() if amount_in == None else amount_in
        tokens = lp.factory.token_from_exchange[lp.name]  

        if(lp.version == UniswapExchangeData.VERSION_V2):

            res0 = lp.get_reserve(tokens[lp.token0])
            res1 = lp.get_reserve(tokens[lp.token1])
            tot_liq = lp.get_liquidity()
            
            if(token_in.token_name == lp.token0):
                liq = amount_in*tot_liq/res0
                amount1 = liq*res1/tot_liq
                amount0, amount1 = lp.remove_liquidity(user_nm, liq, amount_in, amount1) 
            elif(token_in.token_name == lp.token1): 
                liq = amount_in*tot_liq/res1
                amount0 = liq*res0/tot_liq
                amount0, amount1 = lp.remove_liquidity(user_nm, liq, amount0, amount_in) 

        elif(lp.version == UniswapExchangeData.VERSION_V3):  

            tot_liq = lp.get_liquidity()
            sqrt_P = lp.slot0.sqrtPriceX96/2**96

            if(token_in.token_name == lp.token0):
                liq = amount_in*sqrt_P
                (_, _, _, _, amount0, amount1) = lp.burn(user_nm, lwr_tick, upr_tick, liq)
            elif(token_in.token_name == lp.token1): 
                liq = amount_in/sqrt_P
                (_, _, _, _, amount0, amount1) = lp.burn(user_nm, lwr_tick, upr_tick, liq)        
        
            
        return {lp.token0:amount0, lp.token1:amount1}    
    
 
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from .LPQuote import LPQuote

class IndexTokenQuote():
    
    """ 
        Index token quotes
        
        Parameters
        -----------------
        quote_native_tokens : boolean
            Quote LP amount of base pool, otherwise quote indexed pool      
    """        
    
    def __init__(self, lwr_tick = None, upr_tick = None, quote_native_tokens = True):
        self.quote_native_tokens = quote_native_tokens    
        self.lwr_tick = lwr_tick
        self.upr_tick = upr_tick        

    def get_x(self, lp, amt_lp): 
        
        """ get_x

            Given an amount of LP holdings, get x reserve of CPT pair (ie, xy = k)
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            amt_lp : float
                Amount of LP holdings                   

            Returns
            -----------------
            reserve: float
                Amount of x reserve from CPT pair     
        """          
        
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]  
        if(tkn_x.type == 'standard'):
            x_amt = LPQuote(False).get_amount_from_lp(lp, tkn_x, amt_lp, self.lwr_tick, self.upr_tick)
        else: 
            parent_x_tkn = tkn_x.parent_tkn
            parent_lp = self.get_base_lp(lp, tkn_x)
            itkn_lp = LPQuote(False).get_amount_from_lp(lp, tkn_x, amt_lp)
            x_amt = LPQuote(False).get_amount_from_lp(parent_lp, parent_x_tkn, itkn_lp, self.lwr_tick, self.upr_tick)        
        return self.get_native_x(lp, tkn_x, x_amt) if self.quote_native_tokens else x_amt           
        
    def get_y(self, lp, amt_lp): 
        
        """ get_y

            Given an amount of LP holdings, get y reserve of CPT pair (ie, x*y = k)
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap liquidity pool (LP) exchange    
            amt_lp : float
                Amount of LP holdings                   

            Returns
            -----------------
            reserve: float
                Amount of y reserve from CPT pair     
        """        
    
        tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]  
        if(tkn_y.type == 'standard'):
            y_amt = LPQuote(False).get_amount_from_lp(lp, tkn_y, amt_lp, self.lwr_tick, self.upr_tick)
        else:           
            parent_y_tkn = tkn_y.parent_tkn
            parent_lp = self.get_base_lp(lp, tkn_y)
            itkn_lp = LPQuote(False).get_amount_from_lp(lp, tkn_y, amt_lp, self.lwr_tick, self.upr_tick)
            y_amt = LPQuote(False).get_amount_from_lp(parent_lp, parent_y_tkn, itkn_lp, self.lwr_tick, self.upr_tick)        

        return self.get_native_y(lp, tkn_y, y_amt) if self.quote_native_tokens else y_amt     
   
    def base_x_asset(self, lp):
        
        """ base_x_asset

            Get x base asset in x*y CPT pairing
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap liquidity pool (LP) exchange   
               
            Returns
            -----------------
            token: ERC20
                x base asset from CPT pair    
        """          
        
        tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        if(tkn_x.type == 'index'):
            tkn = tkn_x.parent_tkn
        else:
            tkn = tkn_x

        return tkn    

    def base_y_asset(self, lp):
        
        """ base_y_asset

            Get y base asset in x*y CPT pairing
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap liquidity pool (LP) exchange   
               
            Returns
            -----------------
            token: ERC20
                y base asset from CPT pair     
        """                  
        
        tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]
        if(tkn_y.type == 'index'):
            tkn = tkn_y.parent_tkn
        else:
            tkn = tkn_y

        return tkn   

    def get_native_x(self, lp, tkn_x, x_amt):
        
        """ get_native_x

            Get x base asset amount in x*y CPT pairing
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap liquidity pool (LP) exchange   
            tkn_x : ERC20
                ERC20 x token in x*y CPT pairing   
            x_amt : float
                Amount of x token                    
                
            Returns
            -----------------
            reserve: float
                Amount of x reserve from CPT pair     
        """            
        
        parent_lp = self.get_base_lp(lp, tkn_x)
        parent_lp_x_tkn = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token0]
        parent_x_tkn = tkn_x.parent_tkn if tkn_x.type == 'index' else tkn_x

        if(parent_x_tkn.token_name != parent_lp_x_tkn.token_name):
            x_amt = LPQuote().get_amount(parent_lp, parent_x_tkn, x_amt, self.lwr_tick, self.upr_tick) 

        return x_amt

    def get_native_y(self, lp, tkn_y, y_amt):
        
        """ get_native_y

            Get y base asset amount in x*y CPT pairing
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap liquidity pool (LP) exchange   
            tkn_y : ERC20
                ERC20 y token in x*y CPT pairing   
            y_amt : float
                Amount of x token                        
                
            Returns
            -----------------
            reserve: float
                Amount of y reserve from CPT pair     
        """           
        
        parent_lp = self.get_base_lp(lp, tkn_y)
        parent_lp_y_tkn = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token1]
        parent_y_tkn = tkn_y.parent_tkn if tkn_y.type == 'index' else tkn_y

        if(parent_y_tkn.token_name != parent_lp_y_tkn.token_name):
            y_amt = LPQuote().get_amount(parent_lp, parent_y_tkn, y_amt, self.lwr_tick, self.upr_tick) 

        return y_amt
    
    
    def get_base_lp(self, lp, tkn):
        
        """ get_base_lp

            Return the parent LP, given a child LP and one of its reserve tokens, 
                
            Parameters
            -----------------
            lp : UniswapExchange
                Child Uniswap liquidity pool (LP) exchange   
            tkn : ERC20
                Child ERC20 token in x*y CPT pairing                           
                
            Returns
            -----------------
            lp: UniswapExchange
                Parent liquidity pool (LP) exchange  
        """         
        
        if(tkn.type == 'index'):
            base_lp = tkn.parent_lp
        else:
            base_lp = lp.factory.parent_lp
        return lp if base_lp == None else base_lp          
        
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np

class BrownianModel():
    
    def __init__(self, x0=0):
        self.__x0 = float(x0)
    
    def gen_gbm(self, mu, sigma, n_step, T = 1):
        
        if n_step < 30:
            print("WARNING! The number of steps is small. It may not generate a good stochastic process sequence!")
                
        dt = float(T) / n_step
        path = np.zeros(n_step + 1, np.float64)
        path[0] = self.__x0
        for t in range(1, n_step + 1):
            rand = np.random.standard_normal()
            path[t] = path[t - 1] * np.exp((mu - 0.5 * sigma ** 2) * dt + sigma * np.sqrt(dt) * rand)
        return path
    
    
    def gen_gbms(self, mu, sigma, n_step, n_paths = 1, T = 1):
        paths = np.zeros((n_step+1, n_paths), np.float64)
        for k in range(n_paths):
            paths[:,k] = self.gen_gbm(mu, sigma, n_step, T)
        return paths      
    
    def gen_random_walk(self, n_step=100):
        
        # Warning about the small number of steps
        if n_step < 30:
            print("WARNING! The number of steps is small. It may not generate a good stochastic process sequence!")
        
        w = np.ones(n_step)*self.x0
        
        for i in range(1,n_step):
            # Sampling from the Normal distribution with probability 1/2
            yi = np.random.choice([1,-1])
            # Weiner process
            w[i] = w[i-1]+(yi/np.sqrt(n_step))
        
        return w
    
    def gen_normal(self, n_step=100):
        if n_step < 30:
            print("WARNING! The number of steps is small. It may not generate a good stochastic process sequence!")
        
        w = np.ones(n_step)*self.x0
        
        for i in range(1,n_step):
            # Sampling from the Normal distribution
            yi = np.random.normal()
            # Weiner process
            w[i] = w[i-1]+(yi/np.sqrt(n_step))
        
        return w
    


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np
import queue

class ModelQueue():

    def __init__(self):
        self.__model_queue = queue.Queue()
      
    def size(self):
        return self.__model_queue.qsize()
    
    def apply(self, arr, n_points = None):     
        n_points = len(arr) if n_points == None else n_points  
        for k in range(n_points):
            self.__model_queue.put(arr[k])
            
        return self.__model_queue   


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from abc import *

class IExchange(ABC):
    
    @abstractmethod        
    def summary(self, agents):
        pass
    
    @abstractmethod        
    def get_price(self, agents):
        pass
    
    @abstractmethod        
    def get_reserve(self, agents):
        pass    


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass

@dataclass
class FactoryData:
    token_from_exchange: {}
    parent_lp: None
    name: str = None
    address: str = None
    
   


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Original MIT licenced SqrtPriceMath file from chainflip-io
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

from .Shared import *
import math
from . import SafeMath, FullMath


### @title defs based on Q64.96 sqrt price and liquidity
### @notice Contains the math that uses square root of price as a Q64.96 and liquidity to compute deltas

### @notice Gets the next sqrt price given a delta of token0
### @dev Always rounds up, because in the exact output case (increasing price) we need to move the price at least
### far enough to get the desired output amount, and in the exact input case (decreasing price) we need to move the
### price less in order to not send too much output.
### The most precise formula for this is liquidity * sqrtPX96 / (liquidity +- amount * sqrtPX96),
### if this is impossible because of overflow, we calculate liquidity / (liquidity / sqrtPX96 +- amount).
### @param sqrtPX96 The starting price, i.e. before accounting for the token0 delta
### @param liquidity The amount of usable liquidity
### @param amount How much of token0 to add or remove from virtual reserves
### @param add Whether to add or remove the amount of token0
### @return The price after adding or removing amount, depending on add
def getNextSqrtPriceFromAmount0RoundingUp(sqrtPX96, liquidity, amount, add):
    checkInputTypes(uint160=sqrtPX96, uint128=liquidity, uint256=amount, bool=add)
    ## we short circuit amount == 0 because the result is otherwise not guaranteed to equal the input price
    if amount == 0:
        return sqrtPX96
    numerator1 = liquidity << FixedPoint96_RESOLUTION

    if add:
        product = amount * sqrtPX96
        # Modified overflow check compared so to Solidity - product might overflow
        if product < MAX_UINT256:
            ##if product / amount == sqrtPX96:
            denominator = numerator1 + product
            if denominator >= numerator1:
                # Result should be casted into UINT160 without overflowing
                # Adding assert to detect wrong behaviour
                result = FullMath.mulDivRoundingUp(numerator1, sqrtPX96, denominator)
                # result = math.ceil((numerator1 * sqrtPX96) / denominator)
                checkUInt160(result)
                return result
        result = math.ceil(numerator1 / SafeMath.add((numerator1 // sqrtPX96), amount))
        # Adding assert to detect wrong behaviour
        assert result <= MAX_UINT160, "Overflow when casting to UINT160"
        return result

    else:
        ## if the product overflows, we know the denominator underflows
        ## in addition, we must check that the denominator does not underflow
        product = amount * sqrtPX96
        assert product < MAX_UINT256 and numerator1 > product
        denominator = numerator1 - product
        result = FullMath.mulDivRoundingUp(numerator1, sqrtPX96, denominator)
        assert result <= MAX_UINT160, "Overflow when casting to UINT160"
        return result


### @notice Gets the next sqrt price given a delta of token1
### @dev Always rounds down, because in the exact output case (decreasing price) we need to move the price at least
### far enough to get the desired output amount, and in the exact input case (increasing price) we need to move the
### price less in order to not send too much output.
### The formula we compute is within <1 wei of the lossless version: sqrtPX96 +- amount / liquidity
### @param sqrtPX96 The starting price, i.e., before accounting for the token1 delta
### @param liquidity The amount of usable liquidity
### @param amount How much of token1 to add, or remove, from virtual reserves
### @param add Whether to add, or remove, the amount of token1
### @return The price after adding or removing `amount`
def getNextSqrtPriceFromAmount1RoundingDown(sqrtPX96, liquidity, amount, add):
    checkInputTypes(uint160=sqrtPX96, uint128=liquidity, uint256=amount, bool=add)
    ## if we're adding (subtracting), rounding down requires rounding the quotient down (up)
    ## in both cases, avoid a mulDiv for most inputs
    if add:
        quotient = (
            (amount << FixedPoint96_RESOLUTION) // liquidity
            if amount <= MAX_UINT160
            else FullMath.mulDiv(amount, FixedPoint96_Q96, liquidity)
        )

        result = sqrtPX96 + quotient

        checkUInt160(result)
        return result
    else:
        quotient = (
            FullMath.divRoundingUp(amount << FixedPoint96_RESOLUTION, liquidity)
            if amount <= MAX_UINT160
            else FullMath.mulDivRoundingUp(amount, FixedPoint96_Q96, liquidity)
        )

        assert sqrtPX96 > quotient
        ## always fits 160 bits
        result = sqrtPX96 - quotient

        checkUInt160(result)
        return result


### @notice Gets the next sqrt price given an input amount of token0 or token1
### @dev Throws if price or liquidity are 0, or if the next price is out of bounds
### @param sqrtPX96 The starting price, i.e., before accounting for the input amount
### @param liquidity The amount of usable liquidity
### @param amountIn How much of token0, or token1, is being swapped in
### @param zeroForOne Whether the amount in is token0 or token1
### @return sqrtQX96 The price after adding the input amount to token0 or token1
def getNextSqrtPriceFromInput(sqrtPX96, liquidity, amountIn, zeroForOne):
    checkInputTypes(
        uint160=sqrtPX96, uint128=liquidity, uint256=amountIn, bool=zeroForOne
    )
    assert sqrtPX96 > 0
    assert liquidity > 0

    ## round to make sure that we don't pass the target price
    return (
        getNextSqrtPriceFromAmount0RoundingUp(sqrtPX96, liquidity, amountIn, True)
        if zeroForOne
        else getNextSqrtPriceFromAmount1RoundingDown(
            sqrtPX96, liquidity, amountIn, True
        )
    )


### @notice Gets the next sqrt price given an output amount of token0 or token1
### @dev Throws if price or liquidity are 0 or the next price is out of bounds
### @param sqrtPX96 The starting price before accounting for the output amount
### @param liquidity The amount of usable liquidity
### @param amountOut How much of token0, or token1, is being swapped out
### @param zeroForOne Whether the amount out is token0 or token1
### @return sqrtQX96 The price after removing the output amount of token0 or token1
def getNextSqrtPriceFromOutput(sqrtPX96, liquidity, amountOut, zeroForOne):
    checkInputTypes(
        uint160=sqrtPX96, uint128=liquidity, uint256=amountOut, bool=zeroForOne
    )
    assert sqrtPX96 > 0
    assert liquidity > 0

    ## round to make sure that we pass the target price
    return (
        getNextSqrtPriceFromAmount1RoundingDown(sqrtPX96, liquidity, amountOut, False)
        if zeroForOne
        else getNextSqrtPriceFromAmount0RoundingUp(
            sqrtPX96, liquidity, amountOut, False
        )
    )


### @notice Gets the amount0 delta between two prices
### @dev Calculates liquidity / sqrt(lower) - liquidity / sqrt(upper),
### i.e. liquidity * (sqrt(upper) - sqrt(lower)) / (sqrt(upper) * sqrt(lower))
### @param sqrtRatioAX96 A sqrt price
### @param sqrtRatioBX96 Another sqrt price
### @param liquidity The amount of usable liquidity
### @return amount0 Amount of token0 required to cover a position of size liquidity between the two passed prices
def getAmount0Delta(sqrtRatioAX96, sqrtRatioBX96, liquidity, roundUp):

    #print('Delta X')
    
    checkInputTypes(
        uint160=(sqrtRatioAX96, sqrtRatioBX96), uint128=liquidity, bool=roundUp
    )
    if sqrtRatioAX96 > sqrtRatioBX96:
        (sqrtRatioAX96, sqrtRatioBX96) = (sqrtRatioBX96, sqrtRatioAX96)

    numerator1 = liquidity << FixedPoint96_RESOLUTION
    numerator2 = sqrtRatioBX96 - sqrtRatioAX96
    assert numerator2 >= 0

    assert sqrtRatioAX96 > 0
    if roundUp:
        return FullMath.divRoundingUp(
            FullMath.mulDivRoundingUp(numerator1, numerator2, sqrtRatioBX96),
            sqrtRatioAX96,
        )
    else:
        return FullMath.mulDiv(numerator1, numerator2, sqrtRatioBX96) // sqrtRatioAX96


### @notice Gets the amount1 delta between two prices
### @dev Calculates liquidity * (sqrt(upper) - sqrt(lower))
### @param sqrtRatioAX96 A sqrt price
### @param sqrtRatioBX96 Another sqrt price
### @param liquidity The amount of usable liquidity
### @return amount1 Amount of token1 required to cover a position of size liquidity between the two passed prices
def getAmount1Delta(sqrtRatioAX96, sqrtRatioBX96, liquidity, roundUp):

    #print('Delta Y')
    
    checkInputTypes(
        uint160=(sqrtRatioAX96, sqrtRatioBX96), uint128=liquidity, bool=roundUp
    )
    if sqrtRatioAX96 > sqrtRatioBX96:
        (sqrtRatioAX96, sqrtRatioBX96) = (sqrtRatioBX96, sqrtRatioAX96)

    if roundUp:
        return FullMath.mulDivRoundingUp(
            liquidity, (sqrtRatioBX96 - sqrtRatioAX96), FixedPoint96_Q96
        )
    else:
        return FullMath.mulDiv(
            liquidity, (sqrtRatioBX96 - sqrtRatioAX96), FixedPoint96_Q96
        )


### @notice Helper that gets signed token0 delta
### @param sqrtRatioAX96 A sqrt price
### @param sqrtRatioBX96 Another sqrt price
### @param liquidity The change in liquidity for which to compute the amount0 delta
### @return amount0 Amount of token0 corresponding to the passed liquidityDelta between the two prices
def getAmount0DeltaHelper(sqrtRatioAX96, sqrtRatioBX96, liquidity):
    checkInputTypes(uint160=(sqrtRatioAX96, sqrtRatioBX96), int128=liquidity)
    if liquidity < 0:
        return -getAmount0Delta(sqrtRatioAX96, sqrtRatioBX96, abs(liquidity), False)
    else:
        return getAmount0Delta(sqrtRatioAX96, sqrtRatioBX96, abs(liquidity), True)


### @notice Helper that gets signed token1 delta
### @param sqrtRatioAX96 A sqrt price
### @param sqrtRatioBX96 Another sqrt price
### @param liquidity The change in liquidity for which to compute the amount1 delta
### @return amount1 Amount of token1 corresponding to the passed liquidityDelta between the two prices
def getAmount1DeltaHelper(sqrtRatioAX96, sqrtRatioBX96, liquidity):
    checkInputTypes(uint160=(sqrtRatioAX96, sqrtRatioBX96), int128=liquidity)
    if liquidity < 0:
        return -getAmount1Delta(sqrtRatioAX96, sqrtRatioBX96, abs(liquidity), False)
    else:
        return getAmount1Delta(sqrtRatioAX96, sqrtRatioBX96, abs(liquidity), True)



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Original MIT licenced SwapMath file from chainflip-io 
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

from . import SqrtPriceMath, FullMath
from .Shared import *

### @title Computes the result of a swap within ticks
### @notice Contains methods for computing the result of a swap within a single tick price range, i.e., a single tick.

### @notice Computes the result of swapping some amount in, or amount out, given the parameters of the swap
### @dev The fee, plus the amount in, will never exceed the amount remaining if the swap's `amountSpecified` is positive
### @param sqrtRatioCurrentX96 The current sqrt price of the pool
### @param sqrtRatioTargetX96 The price that cannot be exceeded, from which the direction of the swap is inferred
### @param liquidity The usable liquidity
### @param amountRemaining How much input or output amount is remaining to be swapped in#out
### @param feePips The fee taken from the input amount, expressed in hundredths of a bip
### @return sqrtRatioNextX96 The price after swapping the amount in#out, not to exceed the price target
### @return amountIn The amount to be swapped in, of either token0 or token1, based on the direction of the swap
### @return amountOut The amount to be received, of either token0 or token1, based on the direction of the swap
### @return feeAmount The amount of input that will be taken as a fee
def computeSwapStep(
    sqrtRatioCurrentX96, sqrtRatioTargetX96, liquidity, amountRemaining, feePips
):
    checkInputTypes(
        uint160=(sqrtRatioCurrentX96, sqrtRatioTargetX96),
        uint128=liquidity,
        int256=amountRemaining,
        uint24=feePips,
    )

    zeroForOne = sqrtRatioCurrentX96 >= sqrtRatioTargetX96

    # exactIn < 0 means exactOut = True
    exactIn = amountRemaining >= 0

    if exactIn:
        amountRemainingLessFee = FullMath.mulDiv(
            amountRemaining, ONE_IN_PIPS - feePips, ONE_IN_PIPS
        )
        amountIn = (
            SqrtPriceMath.getAmount0Delta(
                sqrtRatioTargetX96, sqrtRatioCurrentX96, liquidity, True
            )
            if zeroForOne
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioCurrentX96, sqrtRatioTargetX96, liquidity, True
            )
        )

        if amountRemainingLessFee >= amountIn:
            sqrtRatioNextX96 = sqrtRatioTargetX96
        else:
            sqrtRatioNextX96 = SqrtPriceMath.getNextSqrtPriceFromInput(
                sqrtRatioCurrentX96, liquidity, amountRemainingLessFee, zeroForOne
            )
    else:
        amountOut = (
            SqrtPriceMath.getAmount1Delta(
                sqrtRatioTargetX96, sqrtRatioCurrentX96, liquidity, False
            )
            if zeroForOne
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioCurrentX96, sqrtRatioTargetX96, liquidity, False
            )
        )

        # amountRemaining <= 0
        if abs(amountRemaining) >= amountOut:
            sqrtRatioNextX96 = sqrtRatioTargetX96
        else:
            sqrtRatioNextX96 = SqrtPriceMath.getNextSqrtPriceFromOutput(
                sqrtRatioCurrentX96, liquidity, abs(amountRemaining), zeroForOne
            )

    max = sqrtRatioTargetX96 == sqrtRatioNextX96

    ## get the input#output amounts
    if zeroForOne:
        amountIn = (
            amountIn
            if (max and exactIn)
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioNextX96, sqrtRatioCurrentX96, liquidity, True
            )
        )
        amountOut = (
            amountOut
            if (max and not exactIn)
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioNextX96, sqrtRatioCurrentX96, liquidity, False
            )
        )

    else:
        amountIn = (
            amountIn
            if (max and exactIn)
            else SqrtPriceMath.getAmount1Delta(
                sqrtRatioCurrentX96, sqrtRatioNextX96, liquidity, True
            )
        )
        amountOut = (
            amountOut
            if (max and not exactIn)
            else SqrtPriceMath.getAmount0Delta(
                sqrtRatioCurrentX96, sqrtRatioNextX96, liquidity, False
            )
        )

    ## cap the output amount to not exceed the remaining output amount
    if (not exactIn) and (amountOut > abs(amountRemaining)):
        checkUInt256(-amountRemaining)
        amountOut = abs(amountRemaining)

    if exactIn and sqrtRatioNextX96 != sqrtRatioTargetX96:
        ## we didn't reach the target, so take the remainder of the maximum input as fee
        checkUInt256(amountRemaining)
        feeAmount = abs(amountRemaining) - amountIn
    else:
        feeAmount = FullMath.mulDivRoundingUp(amountIn, feePips, ONE_IN_PIPS - feePips)

    return (sqrtRatioNextX96, amountIn, amountOut, feeAmount)
